@@ -12,17 +12,20 @@
 #include <rpc/rpc.h>
 #include <fcntl.h>
 #include <time.h>
+#ifndef WIN32
 #include <unistd.h>
+#endif
 #include <utime.h>
 #include <errno.h>
 
+#include "backend.h"
 #include "nfs.h"
 #include "attr.h"
 #include "error.h"
 #include "fh.h"
 #include "fh_cache.h"
 #include "daemon.h"
-#include "backend.h"
+#include "Config/exports.h"
 
 /*
  * check whether stat_cache is for a regular file
@@ -102,7 +105,7 @@ pre_op_attr get_pre_cached(void)
 /*
  * compute post-operation attributes given a stat buffer
  */
-post_op_attr get_post_buf(struct stat buf, struct svc_req * req)
+post_op_attr get_post_buf(backend_statstruct buf, struct svc_req * req)
 {
     post_op_attr result;
 
@@ -114,17 +117,21 @@ post_op_attr get_post_buf(struct stat buf, struct svc_req * req)
 	result.post_op_attr_u.attributes.type = NF3BLK;
     else if (S_ISCHR(buf.st_mode))
 	result.post_op_attr_u.attributes.type = NF3CHR;
+#ifdef S_ISLNK
     else if (S_ISLNK(buf.st_mode))
 	result.post_op_attr_u.attributes.type = NF3LNK;
+#endif				       /* S_ISLNK */
+#ifdef S_ISSOCK
     else if (S_ISSOCK(buf.st_mode))
 	result.post_op_attr_u.attributes.type = NF3SOCK;
+#endif				       /* S_ISSOCK */
     else if (S_ISFIFO(buf.st_mode))
 	result.post_op_attr_u.attributes.type = NF3FIFO;
     else
 	result.post_op_attr_u.attributes.type = NF3REG;
 
     /* adapt permissions for executable files */
-    if (S_ISREG(buf.st_mode)) {
+    if (opt_readable_executables && S_ISREG(buf.st_mode)) {
 	if (buf.st_mode & S_IXUSR)
 	    buf.st_mode |= S_IRUSR;
 	if (buf.st_mode & S_IXGRP)
@@ -169,8 +176,23 @@ post_op_attr get_post_buf(struct stat buf, struct svc_req * req)
 	(buf.st_rdev >> 8) & 0xFF;
     result.post_op_attr_u.attributes.rdev.specdata2 = buf.st_rdev & 0xFF;
     result.post_op_attr_u.attributes.fsid = buf.st_dev;
-    result.post_op_attr_u.attributes.fileid =
-	((uint64) buf.st_dev << 32) + buf.st_ino;
+
+    /* If this is a removable export point, we should return the preset fsid
+       for all objects which resides in the same file system as the exported
+       directory */
+    if (exports_opts & OPT_REMOVABLE) {
+	backend_statstruct epbuf;
+	int res;
+
+	res = backend_lstat(export_path, &epbuf);
+	if (res == -1)
+	    return error_attr;
+	if (buf.st_dev == epbuf.st_dev) {
+	    result.post_op_attr_u.attributes.fsid = export_fsid;
+	}
+    }
+
+    result.post_op_attr_u.attributes.fileid = buf.st_ino;
     result.post_op_attr_u.attributes.atime.seconds = buf.st_atime;
     result.post_op_attr_u.attributes.atime.nseconds = 0;
     result.post_op_attr_u.attributes.mtime.seconds = buf.st_mtime;
@@ -184,10 +206,10 @@ post_op_attr get_post_buf(struct stat buf, struct svc_req * req)
 /*
  * lowlevel routine for getting post-operation attributes
  */
-static post_op_attr get_post_ll(const char *path, uint32 dev, uint32 ino,
+static post_op_attr get_post_ll(const char *path, uint32 dev, uint64 ino,
 				struct svc_req *req)
 {
-    struct stat buf;
+    backend_statstruct buf;
     int res;
 
     if (!path)
@@ -241,7 +263,7 @@ post_op_attr get_post_cached(struct svc_req * req)
  *
  * there is no futimes() function in POSIX or Linux
  */
-static nfsstat3 set_time(const char *path, struct stat buf, sattr3 new)
+static nfsstat3 set_time(const char *path, backend_statstruct buf, sattr3 new)
 {
     time_t new_atime, new_mtime;
     struct utimbuf utim;
@@ -285,7 +307,7 @@ static nfsstat3 set_attr_unsafe(const char *path, nfs_fh3 nfh, sattr3 new)
     unfs3_fh_t *fh = (void *) nfh.data.data_val;
     uid_t new_uid;
     gid_t new_gid;
-    struct stat buf;
+    backend_statstruct buf;
     int res;
 
     res = backend_stat(path, &buf);
@@ -341,7 +363,7 @@ nfsstat3 set_attr(const char *path, nfs_fh3 nfh, sattr3 new)
     int res, fd;
     uid_t new_uid;
     gid_t new_gid;
-    struct stat buf;
+    backend_statstruct buf;
 
     res = backend_lstat(path, &buf);
     if (res != 0)
@@ -445,7 +467,7 @@ mode_t create_mode(sattr3 new)
 nfsstat3 atomic_attr(sattr3 attr)
 {
     if ((attr.uid.set_it == TRUE &&
-         attr.uid.set_uid3_u.uid != backend_geteuid()) ||
+	 attr.uid.set_uid3_u.uid != backend_geteuid()) ||
 	(attr.gid.set_it == TRUE &&
 	 attr.gid.set_gid3_u.gid != backend_getegid()) ||
 	(attr.size.set_it == TRUE && attr.size.set_size3_u.size != 0) ||
