@@ -10,11 +10,18 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/socket.h>
 #include <rpc/rpc.h>
 #include <rpc/pmap_clnt.h>
+
+#ifndef WIN32
+#include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#include <syslog.h>
+#else				       /* WIN32 */
+#include <winsock.h>
+#endif				       /* WIN32 */
+
 #include <fcntl.h>
 #include <memory.h>
 #include <signal.h>
@@ -22,9 +29,9 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <syslog.h>
 #include <time.h>
 #include <unistd.h>
+#include <errno.h>
 
 #if HAVE_RPC_SVC_SOC_H == 1
 # include <rpc/svc_soc.h>
@@ -45,13 +52,12 @@
 #define SIG_PF void(*)(int)
 #endif
 
-#define UNFS_NAME "UNFS3 unfsd 0.9.15 (C) 2006, Pascal Schmidt <unfs3-server@ewetel.net>\n"
+#define UNFS_NAME "UNFS3 unfsd 0.9.17 (C) 2006, Pascal Schmidt <unfs3-server@ewetel.net>\n"
 
 /* write verifier */
 writeverf3 wverf;
 
 /* options and default values */
-int opt_expire_writers = FALSE;
 int opt_detach = TRUE;
 char *opt_exports = "/etc/exports";
 int opt_cluster = FALSE;
@@ -61,7 +67,8 @@ unsigned int opt_nfs_port = NFS_PORT;	/* 0 means RPC_ANYSOCK */
 unsigned int opt_mount_port = NFS_PORT;
 int opt_singleuser = FALSE;
 int opt_brute_force = FALSE;
-in_addr_t opt_bind_addr = INADDR_ANY;
+struct in_addr opt_bind_addr;
+int opt_readable_executables = FALSE;
 
 /* Register with portmapper? */
 int opt_portmapper = TRUE;
@@ -108,13 +115,31 @@ short get_port(struct svc_req *rqstp)
     return (svc_getcaller(rqstp->rq_xprt))->sin_port;
 }
 
+/*
+ * return the socket type of the request (SOCK_STREAM or SOCK_DGRAM)
+ */
+int get_socket_type(struct svc_req *rqstp)
+{
+    int v;
+    socklen_t l;
+
+    l = sizeof(v);
+
+    if (getsockopt(rqstp->rq_xprt->xp_sock, SOL_SOCKET, SO_TYPE, &v, &l) < 0) {
+	logmsg(LOG_CRIT, "unable to determine socket type");
+	return -1;
+    }
+
+    return v;
+}
+
 /*
  * parse command line options
  */
 static void parse_options(int argc, char **argv)
 {
     int opt = 0;
-    char *optstring = "bcC:de:hl:m:n:pstuw";
+    char *optstring = "bcC:de:hl:m:n:prstuw";
 
     while (opt != -1) {
 	opt = getopt(argc, argv, optstring);
@@ -135,17 +160,20 @@ static void parse_options(int argc, char **argv)
 		opt_detach = FALSE;
 		break;
 	    case 'e':
+#ifndef WIN32
 		if (optarg[0] != '/') {
+		    /* A relative path won't work for re-reading the exports
+		       file on SIGHUP, since we are changing directory */
 		    fprintf(stderr, "Error: relative path to exports file\n");
 		    exit(1);
 		}
+#endif
 		opt_exports = optarg;
 		break;
 	    case 'h':
 		printf(UNFS_NAME);
 		printf("Usage: %s [options]\n", argv[0]);
 		printf("\t-h          display this short option summary\n");
-		printf("\t-w          expire writers from fd cache\n");
 		printf("\t-u          use unprivileged port for services\n");
 		printf("\t-d          do not detach from terminal\n");
 		printf("\t-e <file>   file to use instead of /etc/exports\n");
@@ -162,11 +190,13 @@ static void parse_options(int argc, char **argv)
 		printf("\t-b          enable brute force file searching\n");
 		printf
 		    ("\t-l <addr>   bind to interface with specified address\n");
+		printf
+		    ("\t-r          report unreadable executables as readable\n");
 		exit(0);
 		break;
 	    case 'l':
-		opt_bind_addr = inet_addr(optarg);
-		if (opt_bind_addr == (unsigned) -1) {
+		opt_bind_addr.s_addr = inet_addr(optarg);
+		if (opt_bind_addr.s_addr == (unsigned) -1) {
 		    fprintf(stderr, "Invalid bind address\n");
 		    exit(1);
 		}
@@ -188,14 +218,19 @@ static void parse_options(int argc, char **argv)
 	    case 'p':
 		opt_portmapper = FALSE;
 		break;
+	    case 'r':
+		opt_readable_executables = TRUE;
+		break;
 	    case 's':
 		opt_singleuser = TRUE;
+#ifndef WIN32
 		if (backend_getuid() == 0) {
 		    logmsg(LOG_WARNING,
 			   "Warning: running as root with -s is dangerous");
 		    logmsg(LOG_WARNING,
 			   "All clients will have root access to all exported files!");
 		}
+#endif
 		break;
 	    case 't':
 		opt_tcponly = TRUE;
@@ -204,9 +239,6 @@ static void parse_options(int argc, char **argv)
 		opt_nfs_port = 0;
 		opt_mount_port = 0;
 		break;
-	    case 'w':
-		opt_expire_writers = TRUE;
-		break;
 	    case '?':
 		exit(1);
 		break;
@@ -219,6 +251,7 @@ static void parse_options(int argc, char **argv)
  */
 void daemon_exit(int error)
 {
+#ifndef WIN32
     if (error == SIGHUP) {
 	get_squash_ids();
 	exports_parse();
@@ -236,6 +269,7 @@ void daemon_exit(int error)
 	       fd_cache_readers, fd_cache_writers);
 	return;
     }
+#endif				       /* WIN32 */
 
     if (opt_portmapper) {
 	svc_unregister(MOUNTPROG, MOUNTVERS1);
@@ -457,13 +491,11 @@ static void nfs3_program_3(struct svc_req *rqstp, register SVCXPRT * transp)
     if (result != NULL &&
 	!svc_sendreply(transp, (xdrproc_t) _xdr_result, result)) {
 	svcerr_systemerr(transp);
-	logmsg(LOG_CRIT, "unable to send RPC reply, aborting");
-	daemon_exit(CRISIS);
+	logmsg(LOG_CRIT, "unable to send RPC reply");
     }
     if (!svc_freeargs
 	(transp, (xdrproc_t) _xdr_argument, (caddr_t) & argument)) {
 	logmsg(LOG_CRIT, "unable to free XDR arguments");
-	daemon_exit(CRISIS);
     }
     return;
 }
@@ -537,11 +569,11 @@ static void mountprog_3(struct svc_req *rqstp, register SVCXPRT * transp)
     if (result != NULL &&
 	!svc_sendreply(transp, (xdrproc_t) _xdr_result, result)) {
 	svcerr_systemerr(transp);
+	logmsg(LOG_CRIT, "unable to send RPC reply");
     }
     if (!svc_freeargs
 	(transp, (xdrproc_t) _xdr_argument, (caddr_t) & argument)) {
 	logmsg(LOG_CRIT, "unable to free XDR arguments");
-	daemon_exit(CRISIS);
     }
     return;
 }
@@ -635,17 +667,17 @@ static SVCXPRT *create_udp_transport(unsigned int port)
     else {
 	sin.sin_family = AF_INET;
 	sin.sin_port = htons(port);
-	sin.sin_addr.s_addr = opt_bind_addr;
+	sin.sin_addr.s_addr = opt_bind_addr.s_addr;
 	sock = socket(PF_INET, SOCK_DGRAM, 0);
 	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));
 	if (bind(sock, (struct sockaddr *) &sin, sizeof(struct sockaddr))) {
 	    perror("bind");
-	    fprintf(stderr, "Couldn't bind to port %d\n", port);
+	    fprintf(stderr, "Couldn't bind to udp port %d\n", port);
 	    exit(1);
 	}
     }
 
-    transp = svcudp_create(sock);
+    transp = svcudp_bufcreate(sock, NFS_MAX_UDP_PACKET, NFS_MAX_UDP_PACKET);
 
     if (transp == NULL) {
 	fprintf(stderr, "%s\n", "cannot create udp service.");
@@ -667,12 +699,12 @@ static SVCXPRT *create_tcp_transport(unsigned int port)
     else {
 	sin.sin_family = AF_INET;
 	sin.sin_port = htons(port);
-	sin.sin_addr.s_addr = opt_bind_addr;
+	sin.sin_addr.s_addr = opt_bind_addr.s_addr;
 	sock = socket(PF_INET, SOCK_STREAM, 0);
 	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));
 	if (bind(sock, (struct sockaddr *) &sin, sizeof(struct sockaddr))) {
 	    perror("bind");
-	    fprintf(stderr, "Couldn't bind to port %d\n", port);
+	    fprintf(stderr, "Couldn't bind to tcp port %d\n", port);
 	    exit(1);
 	}
     }
@@ -687,6 +719,49 @@ static SVCXPRT *create_tcp_transport(unsigned int port)
     return transp;
 }
 
+/* Run RPC service. This is our own implementation of svc_run(), which
+   allows us to handle other events as well. */
+static void unfs3_svc_run()
+{
+    fd_set readfds;
+    struct timeval tv;
+
+    for (;;) {
+	fd_cache_close_inactive();
+	readfds = svc_fdset;
+	tv.tv_sec = 1;
+	tv.tv_usec = 0;
+	/* Note: On Windows, it's not possible to call select with all sets
+	   empty; to use it as a sleep function. In our case, however,
+	   readfds should never be empty, since we always have our listen
+	   socket. Well, at least hope that our Windows RPC library works
+	   like that. oncrpc-ms does. */
+	switch (select(FD_SETSIZE, &readfds, NULL, NULL, &tv)) {
+	    case -1:
+		if (errno == EINTR) {
+		    continue;
+		}
+		perror("unfs3_svc_run: select failed");
+		return;
+	    case 0:
+		/* timeout */
+		continue;
+	    default:
+		svc_getreqset(&readfds);
+	}
+    }
+}
+
+/*
+ * Generate write verifier based on PID and current time
+ */
+void regenerate_write_verifier(void)
+{
+    *(wverf + 0) = (uint32) getpid();
+    *(wverf + 0) ^= rand();
+    *(wverf + 4) = (uint32) time(NULL);
+}
+
 /*
  * NFSD main function
  * originally generated by rpcgen
@@ -696,10 +771,15 @@ int main(int argc, char **argv)
 {
     register SVCXPRT *tcptransp = NULL, *udptransp = NULL;
     pid_t pid = 0;
+
+#ifndef WIN32
     struct sigaction act;
     sigset_t actset;
+#endif				       /* WIN32 */
     int res;
 
+    opt_bind_addr.s_addr = INADDR_ANY;
+
     parse_options(argc, argv);
     if (optind < argc) {
 	fprintf(stderr, "Error: extra arguments on command line\n");
@@ -707,8 +787,13 @@ int main(int argc, char **argv)
     }
 
     /* init write verifier */
-    *(wverf + 0) = (uint32) getpid();
-    *(wverf + 4) = (uint32) time(NULL);
+    regenerate_write_verifier();
+
+    res = backend_init();
+    if (res == -1) {
+	fprintf(stderr, "backend initialization failed\n");
+	daemon_exit(0);
+    }
 
     if (opt_detach) {
 	/* prepare syslog access */
@@ -735,6 +820,7 @@ int main(int argc, char **argv)
 
     register_mount_service(udptransp, tcptransp);
 
+#ifndef WIN32
     if (opt_detach) {
 	pid = fork();
 	if (pid == -1) {
@@ -742,14 +828,10 @@ int main(int argc, char **argv)
 	    daemon_exit(0);
 	}
     }
+#endif				       /* WIN32 */
 
     if (!opt_detach || pid == 0) {
-	res = backend_init();
-	if (res == -1) {
-	    fprintf(stderr, "backend initialization failed\n");
-	    daemon_exit(0);
-	}
-
+#ifndef WIN32
 	sigemptyset(&actset);
 	act.sa_handler = daemon_exit;
 	act.sa_mask = actset;
@@ -769,9 +851,6 @@ int main(int argc, char **argv)
 	/* don't make directory we started in busy */
 	chdir("/");
 
-	/* no umask to not screw up create modes */
-	umask(0);
-
 	/* detach from terminal */
 	if (opt_detach) {
 	    setsid();
@@ -779,6 +858,10 @@ int main(int argc, char **argv)
 	    fclose(stdout);
 	    fclose(stderr);
 	}
+#endif				       /* WIN32 */
+
+	/* no umask to not screw up create modes */
+	umask(0);
 
 	/* initialize internal stuff */
 	fh_cache_init();
@@ -786,7 +869,7 @@ int main(int argc, char **argv)
 	get_squash_ids();
 	exports_parse();
 
-	svc_run();
+	unfs3_svc_run();
 	exit(1);
 	/* NOTREACHED */
     }
