@@ -16,6 +16,7 @@
 
 #include "nfs.h"
 #include "error.h"
+#include "backend.h"
 
 static int is_stale(void)
 {
@@ -38,6 +39,12 @@ nfsstat3 symlink_err(void)
 	return NFS3ERR_EXIST;
     else if (errno == ENOSPC)
 	return NFS3ERR_NOSPC;
+#ifdef EDQUOT
+    else if (errno == EDQUOT)
+	return NFS3ERR_DQUOT;
+#endif
+    else if (errno == ENOSYS)
+	return NFS3ERR_NOTSUPP;
     else
 	return NFS3ERR_IO;
 }
@@ -58,6 +65,10 @@ nfsstat3 link_err(void)
 	return NFS3ERR_XDEV;
     else if (errno == EMLINK)
 	return NFS3ERR_MLINK;
+#ifdef EDQUOT
+    else if (errno == EDQUOT)
+	return NFS3ERR_DQUOT;
+#endif
     else
 	return symlink_err();
 }
@@ -84,6 +95,8 @@ nfsstat3 readlink_err(void)
 	return NFS3ERR_INVAL;
     else if (errno == EACCES)
 	return NFS3ERR_ACCES;
+    else if (errno == ENOSYS)
+	return NFS3ERR_NOTSUPP;
     else if (is_stale())
 	return NFS3ERR_STALE;
     else
@@ -124,6 +137,10 @@ nfsstat3 write_write_err(void)
 	return NFS3ERR_FBIG;
     else if (errno == ENOSPC)
 	return NFS3ERR_NOSPC;
+#ifdef EDQUOT
+    else if (errno == EDQUOT)
+	return NFS3ERR_DQUOT;
+#endif
     else
 	return NFS3ERR_IO;
 }
@@ -140,6 +157,10 @@ nfsstat3 create_err(void)
 	return NFS3ERR_NOSPC;
     else if (errno == EEXIST)
 	return NFS3ERR_EXIST;
+#ifdef EDQUOT
+    else if (errno == EDQUOT)
+	return NFS3ERR_DQUOT;
+#endif
     else
 	return NFS3ERR_IO;
 }
@@ -168,6 +189,10 @@ nfsstat3 rename_err(void)
 	return NFS3ERR_ROFS;
     else if (errno == ENOSPC)
 	return NFS3ERR_NOSPC;
+#ifdef EDQUOT
+    else if (errno == EDQUOT)
+	return NFS3ERR_DQUOT;
+#endif
     else
 	return NFS3ERR_IO;
 }
@@ -204,6 +229,12 @@ nfsstat3 setattr_err(void)
 	return NFS3ERR_STALE;
     else if (errno == EACCES)
 	return NFS3ERR_ACCES;
+#ifdef EDQUOT
+    else if (errno == EDQUOT)
+	return NFS3ERR_DQUOT;
+#endif
+    else if (errno == EINVAL)
+	return NFS3ERR_INVAL;
     else
 	return NFS3ERR_IO;
 }
