@@ -16,8 +16,6 @@
 #if HAVE_SYS_VMOUNT_H == 1
 #include <sys/vmount.h>		       /* AIX */
 #endif
-#include <sys/socket.h>
-#include <sys/un.h>
 #include <rpc/rpc.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -26,6 +24,10 @@
 #include <string.h>
 #include <unistd.h>
 #include <utime.h>
+#ifndef WIN32
+#include <sys/socket.h>
+#include <sys/un.h>
+#endif				       /* WIN32 */
 
 #if HAVE_STATVFS == 1
 # include <sys/statvfs.h>
@@ -53,8 +55,7 @@
  */
 #define PREP(p,f) do {						\
                       unfs3_fh_t *fh = (void *)f.data.data_val; \
-                      backend_setegid(0);                       \
-                      backend_seteuid(0);                       \
+                      switch_to_root();				\
                       p = fh_decomp(f);				\
                       if (exports_options(p, rqstp, NULL, NULL) == -1) { \
                           memset(&result, 0, sizeof(result));	\
@@ -188,7 +189,7 @@ LOOKUP3res *nfsproc3_lookup_3_svc(LOOKUP3args * argp, struct svc_req * rqstp)
     unfs3_fh_t *fh;
     char *path;
     char obj[NFS_MAXPATHLEN];
-    struct stat buf;
+    backend_statstruct buf;
     int res;
     uint32 gen;
 
@@ -248,28 +249,35 @@ ACCESS3res *nfsproc3_access_3_svc(ACCESS3args * argp, struct svc_req * rqstp)
 	    access |= ACCESS3_READ;
 	if (mode & S_IWUSR)
 	    access |= ACCESS3_MODIFY | ACCESS3_EXTEND;
-	if (mode & S_IXUSR)
-	    access |= ACCESS3_EXECUTE | ACCESS3_READ;
-    }
-
-    /* group permissions */
-    if (has_group(st_cache.st_gid, rqstp)) {
+	if (mode & S_IXUSR) {
+	    access |= ACCESS3_EXECUTE;
+	    if (opt_readable_executables)
+		access |= ACCESS3_READ;
+	}
+    } else if (has_group(st_cache.st_gid, rqstp)) {
+	/* group permissions */
 	if (mode & S_IRGRP)
 	    access |= ACCESS3_READ;
 	if (mode & S_IWGRP)
 	    access |= ACCESS3_MODIFY | ACCESS3_EXTEND;
-	if (mode & S_IXGRP)
-	    access |= ACCESS3_EXECUTE | ACCESS3_READ;
+	if (mode & S_IXGRP) {
+	    access |= ACCESS3_EXECUTE;
+	    if (opt_readable_executables)
+		access |= ACCESS3_READ;
+	}
+    } else {
+	/* other permissions */
+	if (mode & S_IROTH)
+	    access |= ACCESS3_READ;
+	if (mode & S_IWOTH)
+	    access |= ACCESS3_MODIFY | ACCESS3_EXTEND;
+	if (mode & S_IXOTH) {
+	    access |= ACCESS3_EXECUTE;
+	    if (opt_readable_executables)
+		access |= ACCESS3_READ;
+	}
     }
 
-    /* other permissions */
-    if (mode & S_IROTH)
-	access |= ACCESS3_READ;
-    if (mode & S_IWOTH)
-	access |= ACCESS3_MODIFY | ACCESS3_EXTEND;
-    if (mode & S_IXOTH)
-	access |= ACCESS3_EXECUTE | ACCESS3_READ;
-
     /* root is allowed everything */
     if (get_uid(rqstp) == 0)
 	access |= ACCESS3_READ | ACCESS3_MODIFY | ACCESS3_EXTEND;
@@ -323,7 +331,13 @@ READ3res *nfsproc3_read_3_svc(READ3args * argp, struct svc_req * rqstp)
     static READ3res result;
     char *path;
     int fd, res;
-    static char buf[NFS_MAXDATA + 1];
+    static char buf[NFS_MAXDATA_TCP + 1];
+    unsigned int maxdata;
+
+    if (get_socket_type(rqstp) == SOCK_STREAM)
+	maxdata = NFS_MAXDATA_TCP;
+    else
+	maxdata = NFS_MAXDATA_UDP;
 
     PREP(path, argp->file);
     result.status = is_reg();
@@ -332,11 +346,11 @@ READ3res *nfsproc3_read_3_svc(READ3args * argp, struct svc_req * rqstp)
     read_executable(rqstp, st_cache);
 
     /* if bigger than rtmax, truncate length */
-    if (argp->count > NFS_MAXDATA)
-	argp->count = NFS_MAXDATA;
+    if (argp->count > maxdata)
+	argp->count = maxdata;
 
     if (result.status == NFS3_OK) {
-	fd = fd_open(path, argp->file, FD_READ);
+	fd = fd_open(path, argp->file, UNFS3_FD_READ, TRUE);
 	if (fd != -1) {
 	    /* read one more to check for eof */
 	    res = backend_pread(fd, buf, argp->count + 1, argp->offset);
@@ -346,9 +360,9 @@ READ3res *nfsproc3_read_3_svc(READ3args * argp, struct svc_req * rqstp)
 
 	    /* close for real when hitting eof */
 	    if (result.READ3res_u.resok.eof)
-		fd_close(fd, FD_READ, FD_CLOSE_REAL);
+		fd_close(fd, UNFS3_FD_READ, FD_CLOSE_REAL);
 	    else {
-		fd_close(fd, FD_READ, FD_CLOSE_VIRT);
+		fd_close(fd, UNFS3_FD_READ, FD_CLOSE_VIRT);
 		res--;
 	    }
 
@@ -380,13 +394,20 @@ WRITE3res *nfsproc3_write_3_svc(WRITE3args * argp, struct svc_req * rqstp)
 {
     static WRITE3res result;
     char *path;
-    int fd, res;
+    int fd, res, res_close;
 
     PREP(path, argp->file);
     result.status = join(is_reg(), exports_rw());
 
     if (result.status == NFS3_OK) {
-	fd = fd_open(path, argp->file, FD_WRITE);
+	/* We allow caching of the fd only for unstable writes. This is to
+	   prevent generating a new write verifier for failed stable writes,
+	   when the fd was not in the cache. Besides, for stable writes, the
+	   fd will be removed from the cache by fd_close() below, so adding
+	   it to and removing it from the cache is just a waste of CPU cycles 
+	 */
+	fd = fd_open(path, argp->file, UNFS3_FD_WRITE,
+		     (argp->stable == UNSTABLE));
 	if (fd != -1) {
 	    res =
 		backend_pwrite(fd, argp->data.data_val, argp->data.data_len,
@@ -394,21 +415,21 @@ WRITE3res *nfsproc3_write_3_svc(WRITE3args * argp, struct svc_req * rqstp)
 
 	    /* close for real if not UNSTABLE write */
 	    if (argp->stable == UNSTABLE)
-		fd_close(fd, FD_WRITE, FD_CLOSE_VIRT);
+		res_close = fd_close(fd, UNFS3_FD_WRITE, FD_CLOSE_VIRT);
 	    else
-		fd_close(fd, FD_WRITE, FD_CLOSE_REAL);
+		res_close = fd_close(fd, UNFS3_FD_WRITE, FD_CLOSE_REAL);
 
 	    /* we always do fsync(), never fdatasync() */
 	    if (argp->stable == DATA_SYNC)
 		argp->stable = FILE_SYNC;
 
-	    if (res != -1) {
+	    if (res != -1 && res_close != -1) {
 		result.WRITE3res_u.resok.count = res;
 		result.WRITE3res_u.resok.committed = argp->stable;
 		memcpy(result.WRITE3res_u.resok.verf, wverf,
 		       NFS3_WRITEVERFSIZE);
 	    } else {
-		/* error during write */
+		/* error during write or close */
 		result.status = write_write_err();
 	    }
 	} else
@@ -439,7 +460,7 @@ static int store_create_verifier(char *obj, createverf3 verf)
 /*
  * check if a create verifier matches
  */
-static int check_create_verifier(struct stat *buf, createverf3 verf)
+static int check_create_verifier(backend_statstruct * buf, createverf3 verf)
 {
     return ((buf->st_atime ==
 	     (verf[0] | verf[1] << 8 | verf[2] << 16 | verf[3] << 24))
@@ -454,7 +475,7 @@ CREATE3res *nfsproc3_create_3_svc(CREATE3args * argp, struct svc_req * rqstp)
     char obj[NFS_MAXPATHLEN];
     sattr3 new_attr;
     int fd = -1, res = -1;
-    struct stat buf;
+    backend_statstruct buf;
     uint32 gen;
     int flags = O_RDWR | O_CREAT | O_TRUNC | O_NONBLOCK;
 
@@ -510,7 +531,7 @@ CREATE3res *nfsproc3_create_3_svc(CREATE3args * argp, struct svc_req * rqstp)
 	    result.status = NFS3ERR_IO;
 	}
 
-    } else {
+    } else if (result.status == NFS3_OK) {
 	/* open() failed */
 	if (argp->how.mode == EXCLUSIVE && errno == EEXIST) {
 	    /* Check if verifier matches */
@@ -637,6 +658,8 @@ SYMLINK3res *nfsproc3_symlink_3_svc(SYMLINK3args * argp,
     return &result;
 }
 
+#ifndef WIN32
+
 /*
  * create Unix socket
  */
@@ -660,6 +683,8 @@ static int mksocket(const char *path, mode_t mode)
     return res;
 }
 
+#endif				       /* WIN32 */
+
 /*
  * check and process arguments to MKNOD procedure
  */
@@ -921,7 +946,7 @@ FSSTAT3res *nfsproc3_fsstat_3_svc(FSSTAT3args * argp, struct svc_req * rqstp)
 {
     static FSSTAT3res result;
     char *path;
-    struct statvfs buf;
+    backend_statvfsstruct buf;
     int res;
 
     PREP(path, argp->fsroot);
@@ -966,24 +991,29 @@ FSINFO3res *nfsproc3_fsinfo_3_svc(FSINFO3args * argp, struct svc_req * rqstp)
 {
     static FSINFO3res result;
     char *path;
+    unsigned int maxdata;
+
+    if (get_socket_type(rqstp) == SOCK_STREAM)
+	maxdata = NFS_MAXDATA_TCP;
+    else
+	maxdata = NFS_MAXDATA_UDP;
 
     PREP(path, argp->fsroot);
 
     result.FSINFO3res_u.resok.obj_attributes = get_post_cached(rqstp);
 
     result.status = NFS3_OK;
-    result.FSINFO3res_u.resok.rtmax = NFS_MAXDATA;
-    result.FSINFO3res_u.resok.rtpref = NFS_MAXDATA;
+    result.FSINFO3res_u.resok.rtmax = maxdata;
+    result.FSINFO3res_u.resok.rtpref = maxdata;
     result.FSINFO3res_u.resok.rtmult = 4096;
-    result.FSINFO3res_u.resok.wtmax = NFS_MAXDATA;
-    result.FSINFO3res_u.resok.wtpref = NFS_MAXDATA;
+    result.FSINFO3res_u.resok.wtmax = maxdata;
+    result.FSINFO3res_u.resok.wtpref = maxdata;
     result.FSINFO3res_u.resok.wtmult = 4096;
     result.FSINFO3res_u.resok.dtpref = 4096;
     result.FSINFO3res_u.resok.maxfilesize = ~0ULL;
     result.FSINFO3res_u.resok.time_delta.seconds = 1;
     result.FSINFO3res_u.resok.time_delta.nseconds = 0;
-    result.FSINFO3res_u.resok.properties =
-	FSF3_LINK | FSF3_SYMLINK | FSF3_HOMOGENEOUS | FSF3_CANSETTIME;
+    result.FSINFO3res_u.resok.properties = backend_fsinfo_properties;
 
     return &result;
 }
@@ -1003,7 +1033,8 @@ PATHCONF3res *nfsproc3_pathconf_3_svc(PATHCONF3args * argp,
     result.PATHCONF3res_u.resok.name_max = NFS_MAXPATHLEN;
     result.PATHCONF3res_u.resok.no_trunc = TRUE;
     result.PATHCONF3res_u.resok.chown_restricted = FALSE;
-    result.PATHCONF3res_u.resok.case_insensitive = FALSE;
+    result.PATHCONF3res_u.resok.case_insensitive =
+	backend_pathconf_case_insensitive;
     result.PATHCONF3res_u.resok.case_preserving = TRUE;
 
     return &result;
@@ -1020,7 +1051,7 @@ COMMIT3res *nfsproc3_commit_3_svc(COMMIT3args * argp, struct svc_req * rqstp)
 
     if (result.status == NFS3_OK) {
 	res = fd_sync(argp->file);
-	if (res != 1)
+	if (res != -1)
 	    memcpy(result.COMMIT3res_u.resok.verf, wverf, NFS3_WRITEVERFSIZE);
 	else
 	    /* error during fsync() or close() */
