@@ -7,13 +7,15 @@
 
 #include "config.h"
 
+#ifndef WIN32
 #include <pwd.h>
+#include <syslog.h>
+#include <unistd.h>
+#endif				       /* WIN32 */
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <rpc/rpc.h>
 #include <stdlib.h>
-#include <syslog.h>
-#include <unistd.h>
 
 #include "nfs.h"
 #include "mount.h"
@@ -34,10 +36,10 @@ static int can_switch = TRUE;
  */
 void get_squash_ids(void)
 {
-    struct passwd *passwd;
+    backend_passwdstruct *passwd;
 
     if (can_switch) {
-	passwd = getpwnam("nobody");
+	passwd = backend_getpwnam("nobody");
 	if (passwd) {
 	    squash_uid = passwd->pw_uid;
 	    squash_gid = passwd->pw_gid;
@@ -69,11 +71,15 @@ static int mangle(int id, int squash)
 int get_uid(struct svc_req *req)
 {
     struct authunix_parms *auth = (void *) req->rq_clntcred;
+    int squash = squash_uid;
+
+    if (exports_anonuid() != ANON_NOTSPECIAL)
+	squash = exports_anonuid();
 
     if (req->rq_cred.oa_flavor == AUTH_UNIX)
-	return mangle(auth->aup_uid, squash_uid);
+	return mangle(auth->aup_uid, squash);
     else
-	return squash_uid;	       /* fallback if no uid given */
+	return squash;		       /* fallback if no uid given */
 }
 
 /*
@@ -82,11 +88,15 @@ int get_uid(struct svc_req *req)
 static int get_gid(struct svc_req *req)
 {
     struct authunix_parms *auth = (void *) req->rq_clntcred;
+    int squash = squash_gid;
+
+    if (exports_anongid() != ANON_NOTSPECIAL)
+	squash = exports_anongid();
 
     if (req->rq_cred.oa_flavor == AUTH_UNIX)
-	return mangle(auth->aup_gid, squash_gid);
+	return mangle(auth->aup_gid, squash);
     else
-	return squash_gid;	       /* fallback if no gid given */
+	return squash;		       /* fallback if no gid given */
 }
 
 /*
@@ -118,6 +128,18 @@ int has_group(int group, struct svc_req *req)
     return FALSE;
 }
 
+/*
+ * switch to root
+ */
+void switch_to_root()
+{
+    if (!can_switch)
+	return;
+
+    backend_setegid(0);
+    backend_seteuid(0);
+}
+
 /*
  * switch user and group id to values listed in request
  */
@@ -156,25 +178,20 @@ void switch_user(struct svc_req *req)
 /*
  * re-switch to root for reading executable files
  */
-void read_executable(struct svc_req *req, struct stat buf)
+void read_executable(struct svc_req *req, backend_statstruct buf)
 {
     int have_exec = 0;
 
-    /* bail out if read permission is there */
-    if (buf.st_mode & S_IROTH)
-	return;
-    else if ((buf.st_mode & S_IRUSR) && is_owner(buf.st_uid, req))
-	return;
-    else if ((buf.st_mode & S_IRGRP) && has_group(buf.st_gid, req))
-	return;
-
-    /* check for execute permission */
-    if (buf.st_mode & S_IXOTH)
-	have_exec = 1;
-    else if ((buf.st_mode & S_IXUSR) && is_owner(buf.st_uid, req))
-	have_exec = 1;
-    else if ((buf.st_mode & S_IXGRP) && has_group(buf.st_gid, req))
-	have_exec = 1;
+    if (is_owner(buf.st_uid, req)) {
+	if (!(buf.st_mode & S_IRUSR) && (buf.st_mode & S_IXUSR))
+	    have_exec = 1;
+    } else if (has_group(buf.st_gid, req)) {
+	if (!(buf.st_mode & S_IRGRP) && (buf.st_mode & S_IXGRP))
+	    have_exec = 1;
+    } else {
+	if (!(buf.st_mode & S_IROTH) && (buf.st_mode & S_IXOTH))
+	    have_exec = 1;
+    }
 
     if (have_exec) {
 	backend_setegid(0);
